"""Plotting functions.

Matthew Alger
The Australian National University
2016
"""

import logging

import astropy.io.fits
import astropy.wcs
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm
import numpy
import sklearn.metrics

from .config import config

ARCMIN = 1 / 60  # deg
FITS_CONVENTION = 1
IMAGE_DIAMETER = config['surveys']['atlas']['fits_width']

wcs = None


def _init_wcs():
    """Initialise the ATLAS image WCS. Sets global variable wcs."""
    with astropy.io.fits.open(config['data_sources']['atlas_image'],
                              ignore_blank=True) as atlas_image:
        global wcs
        wcs = astropy.wcs.WCS(atlas_image[0].header).dropaxis(3).dropaxis(2)


def ra_dec_to_pixels(subject_coords, coords):
    if wcs is None:
        _init_wcs()

    offset, = wcs.all_world2pix([subject_coords], FITS_CONVENTION)
    # The coords are of the middle of the subject.
    coords = wcs.all_world2pix(coords, FITS_CONVENTION)
    coords -= offset
    
    coords[:, 0] /= config['surveys']['atlas']['mosaic_scale_x'] * 424 / 200
    coords[:, 1] /= config['surveys']['atlas']['mosaic_scale_y'] * 424 / 200
    
    coords += [40, 40]
    
    return coords


def plot_points_on_background(points, background, noise=False, base_size=200):
    plt.imshow(background, cmap='gray')

    colours = cm.rainbow(numpy.linspace(0, 1, len(points)))
    for colour, (x, y) in zip(colours, points):
        if noise:
            x += numpy.random.normal(scale=0.5)
            y += numpy.random.normal(scale=0.5)
        plt.scatter(x, y, marker='x', c=colour, s=base_size)
    plt.axis('off')
    plt.xlim((0, background.shape[0]))
    plt.ylim((0, background.shape[1]))


def plot_classifications(atlas_vector, ir_matrix, labels, base_size=200,
                         noise=False):
    image = atlas_vector[2 : 2 + IMAGE_DIAMETER ** 2].reshape(
            (IMAGE_DIAMETER, IMAGE_DIAMETER))[60:140, 60:140]
    radio_coords = atlas_vector[:2]

    nearby = atlas_vector[2 + IMAGE_DIAMETER ** 2:] < ARCMIN
    nearby_labels = labels[nearby]

    ir_coords_ = ir_matrix[nearby, :2][nearby_labels.astype(bool)]
    ir_coords_ = ra_dec_to_pixels(radio_coords, ir_coords_)

    ir_coords = []
    for label, coords in zip(nearby_labels.nonzero()[0], ir_coords_):
        label = nearby_labels[label]
        ir_coords.extend([coords] * label)  # For label multiplicity.
    
    plot_points_on_background(ir_coords, image, base_size=base_size,
                              noise=noise)


def plot_classifications_row(atlas_vector, ir_matrix, classifier_labels,
                             rgz_labels, norris_labels, base_size=200,
                             noise=False):
    plt.subplot(1, 3, 1)
    plt.title('Classifier')
    plot_classifications(atlas_vector, ir_matrix, classifier_labels,
                         base_size=base_size, noise=noise)
    
    plt.subplot(1, 3, 2)
    plt.title('RGZ')
    plot_classifications(atlas_vector, ir_matrix, rgz_labels,
                         base_size=base_size, noise=noise)
    
    plt.subplot(1, 3, 3)
    plt.title('Norris')
    plot_classifications(atlas_vector, ir_matrix, norris_labels,
                         base_size=base_size, noise=noise)


def vertical_scatter(xs, ys, style='bx', rotation='horizontal',
                     line=False, x_tick_offset=0, numeric_x=False):
    """Plots a vertical scatter plot.

    xs: List of x labels.
    ys: List of lists of points to scatter vertically.
    style: Plots point style. Default 'bx'.
    rotation: x label rotation. Default 'horizontal'.
    line: Draw lines between corresponding points. Default False.
    x_tick_offset: How far to offset the x tick labels. Default 0.
    numeric_x: Whether the x labels should be treated as numeric. Default False.
    """
    if not numeric_x:
        for x in range(len(xs)):
            plt.plot([x] * len(ys[x]), ys[x], style)
        if line:
            assert all(len(y) == len(ys[0]) for y in ys)
            ys_t = list(zip(*ys))
            for y in range(len(ys[0])):
                plt.plot(range(len(xs)), ys_t[y])
        plt.xticks([i + x_tick_offset for i in range(len(xs))], xs,
                   rotation=rotation)
        plt.xlim((-0.5, len(xs) - 0.5))  # Adds a little buffer.
    else:
        for xi, x in enumerate(xs):
            plt.plot([float(x)] * len(ys[xi]), ys[xi], style)
        if line:
            assert all(len(y) == len(ys[0]) for y in ys)
            ys_t = list(zip(*ys))
            for y in range(len(ys[0])):
                plt.plot(xs, ys_t[y])


def violinplot(xs, ys, rotation='horizontal', points=100, x_tick_offset=0,
               facecolour='lightgreen', edgecolour='green'):
    """Plots a vertical scatter plot.

    xs: List of x labels.
    ys: List of lists of points to scatter vertically.
    rotation: x label rotation. Default 'horizontal'.
    points: Number of points to use in the density estimate.
    x_tick_offset: How far to offset the x tick labels. Default 0.
    facecolour: Colour of the violin plots. Default light green.
    edgecolour: Colour of the violin lines. Default green.
    """
    vp = plt.violinplot(ys, showmeans=True, showextrema=False, points=points)
    # plt.violinplot has no arguments that let us set colours, so we have to do
    # it ourselves. http://stackoverflow.com/a/26291582/1105803
    for pc in vp['bodies']:
        pc.set_facecolor(facecolour)
        pc.set_edgecolor(edgecolour)
    vp['cmeans'].set_color(edgecolour)
    plt.xticks([1 + i + x_tick_offset for i in range(len(xs))],
               xs, rotation=rotation)
    plt.xlim((0.5, len(xs) + 0.5))  # Adds a little buffer.


def vertical_scatter_ba(results, targets, ylim=(70, 100), violin=False,
                        minorticks=False, percentage=True, **kwargs):
    """Plot a vertical scatter plot of balanced accuracies.

    results: Results object.
    targets: Target labels.
    ylim: (lower, upper) y axis.
    violin: Plot a violin plot instead. Default False.
    minorticks: Use minor ticks. Default False.
    percentage: Plot percentage rather than raw balanced accuracy. Default True.
    kwargs: Keyword arguments passed to vertical_scatter.
    """
    xs = sorted(results.method_idx, key=results.method_idx.get)
    ys = []
    for method in xs:
        y = []
        for split in range(results.n_splits):
            mask = results.get_mask(method, split)
            split_results = results[method, split][mask].round()
            split_targets = targets[mask]
            if len(split_results) == 0:
                continue
            # Calculate balanced accuracy.
            cm = sklearn.metrics.confusion_matrix(split_targets, split_results)
            tp = cm[1, 1]
            n, p = cm.sum(axis=1)
            tn = cm[0, 0]
            ba = (tp / p + tn / n) / 2
            if percentage:
                ba *= 100
            y.append(ba)
        logging.info('Average balanced accuracy ({}): {:.02%}'.format(
                method, numpy.mean(y)))
        logging.info('Standard deviation ({}): {:.02%}'.format(
                method, numpy.std(y)))
        ys.append(y)

    if violin:
        violinplot(xs, ys, **kwargs)
    else:
        vertical_scatter(xs, ys, **kwargs)
    plt.ylim(ylim)
    plt.grid(b=True, which='both', axis='y', color='grey', linestyle='-',
             alpha=0.5)
    if minorticks:
        plt.minorticks_on()
    plt.tick_params(axis='x', which='minor', length=0)
    plt.ylabel('Balanced accuracy' + (' (%)' if percentage else ''))


def fillbetween(xs, ys, facecolour='lightgreen', edgecolour='green',
                marker='x', facealpha=1.0, facekwargs=None, **kwargs):
    """Plots a line plot with error represented by filled-between lines.

    xs: List of x values.
    ys: List of lists of y values.
    facecolour: Colour of the filled-between lines. Default light green.
    edgecolour: Colour of the central line. Default green.
    marker: Point marker. Default 'x'.
    facealpha: Alpha value of filled section. Default 1.0.
    facekwargs: Keyword arguments to pass to fill_between. Default {}.
    """
    facekwargs = facekwargs or {}
    means = numpy.mean(ys, axis=1)
    stds = numpy.std(ys, axis=1)
    plt.plot(xs, means, color=edgecolour, marker=marker, **kwargs)
    plt.fill_between(xs, means - stds, means + stds, color=facecolour,
                     alpha=facealpha, linewidth=0, **facekwargs)
    plt.fill_between(xs, means - stds, means + stds, facecolor='None',
                     edgecolor=edgecolour, alpha=0.5, **facekwargs)



import argparse

import h5py
import sklearn.externals.joblib
import sklearn.linear_model


def main(training_h5_path, classifier_path):
    with h5py.File(training_h5_path, 'r') as training_h5:
        features = training_h5['features']
        labels = training_h5['labels']
        lr = sklearn.linear_model.LogisticRegression(class_weight='balanced')
        lr.fit(features, labels)
        sklearn.externals.joblib.dump(lr, classifier_path)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--training')
    parser.add_argument('--classifier')
    args = parser.parse_args()
    main(args.training, args.classifier)




"""Generates sets of testing indices for the galaxy classification task.

Matthew Alger
The Australian National University
2016
"""

import argparse

import h5py
import numpy

from .config import config


ATLAS_WIDTH = config['surveys']['atlas']['fits_width']
ATLAS_HEIGHT = config['surveys']['atlas']['fits_height']
ATLAS_SIZE = ATLAS_WIDTH * ATLAS_HEIGHT


def get_nearby_galaxies(atlas_vector, radius=1/60):
    """Gets all nearby galaxy IDs.

    atlas_vector: crowdastro ATLAS vector.
    radius: Radius to call a galaxy "nearby", in degrees. Default 1 arcmin.
    -> list of ints
    """
    return (atlas_vector[2 + ATLAS_SIZE:] <= radius).nonzero()[0]


def main(c_h5, t_h5, n, p, p_cnn=0.1, add=False):
    ir_survey = c_h5.attrs['ir_survey']
    ir_survey_ = t_h5.attrs['ir_survey']
    assert ir_survey == ir_survey_
    fan_labels = c_h5['/{}/cdfs/fan_labels'.format(ir_survey)].value
    norris_labels = c_h5['/{}/cdfs/norris_labels'.format(ir_survey)].value

    if '/{}/cdfs/test_sets'.format(ir_survey) in c_h5 and not add:
        raise ValueError('Test sets already exist.')

    candidates = []
    for atlas_id, atlas in enumerate(c_h5['/atlas/cdfs/numeric']):
        # Get all nearby galaxies.
        nearby = get_nearby_galaxies(atlas)
        # If Norris and Fan agree on all nearby galaxies, then add this ATLAS
        # object to a set of candidates.
        if all(fan_labels[n] == norris_labels[n] for n in nearby):
            candidates.append(atlas_id)

    n_atlas = c_h5['/atlas/cdfs/numeric'].shape[0]

    if not add:
        # Pull out p_cnn% of the ATLAS objects for use training the CNN.
        cnn_train_set_atlas = list(range(n_atlas))
        numpy.random.shuffle(cnn_train_set_atlas)
        cnn_train_set_atlas = set(cnn_train_set_atlas[:int(n_atlas * p_cnn)])
        cnn_train_set = []
        for atlas_id in cnn_train_set_atlas:
            nearby = get_nearby_galaxies(c_h5['/atlas/cdfs/numeric'][atlas_id])
            cnn_train_set.extend(nearby)
        cnn_train_set.sort()
        cnn_train_set_ = set(cnn_train_set)
    else:
        cnn_train_set = set(
            c_h5['/{}/cdfs/cnn_train_set'.format(ir_survey)].value)
        cnn_train_set_atlas = set()
        for atlas_id, atlas in enumerate(c_h5['/atlas/cdfs/numeric']):
            nearby = get_nearby_galaxies(atlas)
            for i in nearby:
                if i in cnn_train_set:
                    cnn_train_set_atlas.add(atlas_id)
                    break
        cnn_train_set = sorted(cnn_train_set)
        cnn_train_set_ = set(cnn_train_set)

    # Generate the test sets. Make sure none of the ATLAS objects are in the
    # CNN training set.
    test_sets = []
    if add:
        for test_set in c_h5['/{}/cdfs/test_sets'.format(ir_survey)].value:
            test_sets.append(list(test_set))
        assert len(test_sets) == c_h5['/{}/cdfs/test_sets'.format(ir_survey)
                                      ].shape[0]
    for i in range(n):
        # Select at random, without replacement, candidate ATLAS objects.
        candidates_ = list(set(candidates) - cnn_train_set_atlas)
        numpy.random.shuffle(candidates_)
        atlas_test_set = []
        while len(atlas_test_set) < int(n_atlas * p):
            for atlas_id in candidates_:
                assert atlas_id not in cnn_train_set_atlas
                nearby = get_nearby_galaxies(
                    c_h5['/atlas/cdfs/numeric'][atlas_id])
                for j in nearby:
                    # Check for overlaps.
                    if j in cnn_train_set_:
                        break
                else:
                    # No overlaps!
                    atlas_test_set.append(atlas_id)

        # Get all nearby galaxies and add all nearby galaxies to the test set.
        test_set = []
        for atlas_id in atlas_test_set:
            assert atlas_id not in cnn_train_set_atlas
            nearby = get_nearby_galaxies(c_h5['/atlas/cdfs/numeric'][atlas_id])
            for j in nearby:
                # Have to make sure there's no overlaps here.
                assert j not in cnn_train_set_
            test_set.extend(nearby)

        test_sets.append(list(set(test_set)))

    # Because the test sets are galaxy-based but the drawing was radio-based,
    # we may have unequal length lists, so we'll crop them.
    min_length = min(len(test_set) for test_set in test_sets)
    test_sets_ = []
    for test_set in test_sets:
        # Have to reshuffle so we don't bias against later IDs.
        numpy.random.shuffle(test_set)
        test_sets_.append(sorted(test_set[:min_length]))

    test_sets = numpy.array(test_sets_)
    if add:
        del c_h5['/{}/cdfs/test_sets'.format(ir_survey)]
        del t_h5['test_sets']
    c_h5.create_dataset('/{}/cdfs/test_sets'.format(ir_survey),
                        data=test_sets)
    t_h5.create_dataset('test_sets', data=test_sets)

    if not add:
        cnn_train_set = numpy.array(cnn_train_set)
        c_h5.create_dataset('/{}/cdfs/cnn_train_set'.format(ir_survey),
                            data=cnn_train_set)
        t_h5.create_dataset('cnn_train_set', data=cnn_train_set)
    else:
        assert 'cnn_train_set' in c_h5['/{}/cdfs/'.format(ir_survey)]
        assert 'cnn_train_set' in t_h5


def _populate_parser(parser):
    parser.description = 'Generates sets of testing indices for the galaxy ' \
                         'classification task.'
    parser.add_argument('--crowdastro', default='data/crowdastro.h5',
                        help='Crowdastro HDF5 file')
    parser.add_argument('--training', default='data/training.h5',
                        help='Training HDF5 file')
    parser.add_argument('--n', default=5, type=int,
                        help='Number of test sets')
    parser.add_argument('--p', default=0.5, type=float,
                        help='Percentage size of test sets')
    parser.add_argument('--add', action='store_true',
                        help='Add new test sets to existing test sets')


def _main(args):
    with h5py.File(args.crowdastro, 'r+') as c_h5:
        with h5py.File(args.training, 'r+') as t_h5:
            main(c_h5, t_h5, args.n, args.p, add=args.add)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Shows filters of a convolutional neural network.

Matthew Alger
The Australian National University
2016
"""

import argparse

import h5py
import matplotlib.pyplot as plt
import numpy

from crowdastro.config import config


def main(model_path, weights_path, training_h5_path):
    import keras.models
    with open(model_path, 'r') as model_json:
        model = keras.models.model_from_json(model_json.read())
    model.compile(loss='binary_crossentropy', optimizer='adadelta')
    model.load_weights(weights_path)

    for i, layer in enumerate(model.layers):
        print(i, layer)

    conv_layer_0 = model.layers[0].get_weights()[0]
    conv_layer_1 = model.layers[3].get_weights()[0]
    conv_layer_2 = model.layers[6].get_weights()[0]

    # plt.figure(figsize=[8, 4])
    # for i, filt in enumerate(conv_layer_0):
    #     plt.subplot(4, 8, i + 1)
    #     plt.axis('off')
    #     filt = filt[0]
    #     plt.pcolor(filt, cmap='viridis')
    # plt.subplots_adjust(hspace=0.05, wspace=0.05, left=0.00, bottom=0.00,
    #                     top=1.00, right=1.00)
    # plt.show()

    # plt.figure(figsize=[8, 8])
    # for i, filts in enumerate(conv_layer_1):
    #     for j, filt in enumerate(filts):
    #         plt.subplot(32, 32, i * 32 + j + 1)
    #         plt.axis('off')
    #         plt.pcolor(filt, cmap='viridis')
    # plt.subplots_adjust(hspace=0.05, wspace=0.05, left=0.00, bottom=0.00,
    #                     top=1.00, right=1.00)
    # plt.show()

    # plt.figure(figsize=[8, 8])
    # for i, filts in enumerate(conv_layer_2):
    #     for j, filt in enumerate(filts):
    #         plt.subplot(32, 32, i * 32 + j + 1)
    #         plt.axis('off')
    #         plt.pcolor(filt, cmap='viridis')
    # plt.subplots_adjust(hspace=0.05, wspace=0.05, left=0.00, bottom=0.00,
    #                     top=1.00, right=1.00)
    # plt.show()

    encoded_0 = keras.backend.function([model.layers[0].input],
                                       [model.layers[0].output])
    encoded_1 = keras.backend.function([model.layers[0].input],
                                       [model.layers[3].output])
    encoded_2 = keras.backend.function([model.layers[0].input],
                                       [model.layers[6].output])
    print('Weights shapes:', [i.shape for i in model.get_weights()])
    print([(i, l) for i, l in enumerate(model.layers)])

    with h5py.File(training_h5_path, 'r') as f_h5:
        n_raw_features = config['surveys'][
            f_h5.attrs['ir_survey']]['n_features']
        for i in numpy.random.randint(f_h5['raw_features'].shape[0],
                                      size=2):
            img = f_h5['raw_features'][i, n_raw_features:].reshape(
                (1, 1, 32, 32))

            # # Print shapes out.
            # for j, layer in enumerate(model.layers):
            #     f = keras.backend.function([model.layers[0].input],
            #                                [layer.output])
            #     print(layer)
            #     print(j)
            #     print(f([img])[0].shape)
            out_0 = encoded_0([img])[0].reshape((32, 29, 29))
            out_1 = encoded_1([img])[0].reshape((32, 11, 11))
            out_2 = encoded_2([img])[0].reshape((32, 2, 2))

            plt.figure(figsize=[2 + 4 + 4 + 4, 8])

            ax = plt.subplot2grid((8, 2 + 4 + 4 + 4), (3, 0), colspan=2,
                                  rowspan=2)
            ax.pcolor(img[0, 0], cmap='viridis')
            ax.axis('off')
            for j, convolved in enumerate(out_0):
                ax = plt.subplot2grid((8, 2 + 4 + 4 + 4), (j // 4, 2 + j % 4))
                plt.axis('off')
                plt.pcolor(convolved, cmap='viridis')
            for j, convolved in enumerate(out_1):
                ax = plt.subplot2grid((8, 2 + 4 + 4 + 4), (j // 4, 6 + j % 4))
                plt.axis('off')
                plt.pcolor(convolved, cmap='viridis')
            for j, convolved in enumerate(out_2):
                ax = plt.subplot2grid((8, 2 + 4 + 4 + 4), (j // 4, 10 + j % 4))
                plt.axis('off')
                plt.pcolor(convolved, cmap='viridis')
            plt.show()


def _populate_parser(parser):
    parser.description = 'Shows filters of a convolutional neural network.'
    parser.add_argument('--model', help='path to model JSON',
                        default='data/model.json')
    parser.add_argument('--weights', help='path to weights HDF5',
                        default='data/weights.h5')
    parser.add_argument('--training', help='path to training HDF5',
                        default='data/training.h5')


def _main(args):
    main(args.model, args.weights, args.training)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Finds labels for ELAIS-S1 objects.

The Australian National University
2016
"""

import argparse

import h5py
import sklearn.externals.joblib
import sklearn.linear_model


def main(training_h5_path, classifier_path, output_h5_path):
    with h5py.File(training_h5_path, 'r') as f_h5:
        features = f_h5['features']
        classifier = sklearn.externals.joblib.load(classifier_path)
        predictions = classifier.predict(features)
        with h5py.File(output_h5_path, 'w') as g_h5:
            g_h5.create_dataset('predictions', data=predictions, dtype=bool)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--training', default='data/training.h5')
    parser.add_argument('--classifier', default='data/classifier.dat')
    parser.add_argument('--output', default='data/predictions.h5')
    args = parser.parse_args()
    main(args.training, args.classifier, args.output)




"""Generates galaxy labels for individual annotators.

Matthew Alger
The Australian National University
2016
"""

import argparse

import h5py
import numpy
import sklearn.neighbors

from .config import config

ARCMIN = 1 / 60  # deg
ATLAS_IMAGE_SIZE = (config['surveys']['atlas']['fits_width'] * 
                    config['surveys']['atlas']['fits_height'])


def main(f_h5, ir_survey, overwrite=False):
    """
    f_h5: crowdastro HDF5 file.
    ir_survey: 'wise' or 'swire'.
    overwrite: Whether to overwrite existing annotator labels. Default False.
    """
    usernames = sorted({username for username in
                        f_h5['/atlas/cdfs/classification_usernames'].value
                        if username})
    usernames = {j:i for i, j in enumerate(usernames)}
    n_annotators = len(usernames)
    n_examples = f_h5['/{}/cdfs/numeric'.format(ir_survey)].shape[0]

    if (('/{}/cdfs/rgz_raw_labels'.format(ir_survey) in f_h5
         or '/{}/cdfs/rgz_raw_labels_mask'.format(ir_survey) in f_h5)
         and overwrite):
        del f_h5['/{}/cdfs/rgz_raw_labels'.format(ir_survey)]
        del f_h5['/{}/cdfs/rgz_raw_labels_mask'.format(ir_survey)]

    labels = f_h5.create_dataset(
            '/{}/cdfs/rgz_raw_labels'.format(ir_survey),
            data=numpy.zeros((n_annotators, n_examples)).astype(bool))
    unseen = f_h5.create_dataset(
            '/{}/cdfs/rgz_raw_labels_mask'.format(ir_survey),
            data=numpy.ones((n_annotators, n_examples)).astype(bool))

    ir_tree = sklearn.neighbors.KDTree(
            f_h5['/{}/cdfs/numeric'.format(ir_survey)][:, :2])

    # What objects has each annotator seen? Assume that if they have labelled a
    # radio object, then they have seen everything within 1 arcminute.
    for class_pos, username in zip(
            f_h5['/atlas/cdfs/classification_positions'],
            f_h5['/atlas/cdfs/classification_usernames']):
        if username not in usernames:
            # Skip unidentified users.
            continue

        atlas_id, ra, dec, primary = class_pos
        distances = f_h5['/atlas/cdfs/numeric'][atlas_id, 2 + ATLAS_IMAGE_SIZE:]
        assert distances.shape[0] == n_examples
        nearby = distances <= ARCMIN
        unseen[usernames[username], nearby] = False  # Unmask nearby objects.

        # Now, label the example we are looking at now.
        if numpy.isnan(ra) or numpy.isnan(dec):
            # No IR source, so we can just skip this.
            continue

        # Get the nearest IR object and assign the label.
        ((dist,),), ((ir,),) = ir_tree.query([(ra, dec)])
        if dist > config['surveys'][ir_survey]['distance_cutoff']:
            logging.debug('Click has no corresponding IR.')
            continue

        labels[usernames[username], ir] = 1


def _populate_parser(parser):
    parser.description = 'Generates galaxy labels for individual annotators.'
    parser.add_argument('--h5', default='data/crowdastro.h5',
                        help='HDF5 IO file')
    parser.add_argument('--overwrite', action='store_true',
                        help='overwrite existing annotator labels')


def _main(args):
    with h5py.File(args.h5, 'r+') as f_h5:
        ir_survey = f_h5.attrs['ir_survey']
        main(f_h5, ir_survey, overwrite=args.overwrite)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Command-line interface for crowdastro.

Matthew Alger
The Australian National University
2016
"""

import argparse
import logging
import sys

from crowdastro import __description__
from crowdastro import __version__
from crowdastro import compile_cnn
from crowdastro import consensuses
from crowdastro import generate_annotator_labels
from crowdastro import generate_cnn_outputs
from crowdastro import generate_dataset
from crowdastro import generate_test_sets
from crowdastro import generate_training_data
from crowdastro import import_data
from crowdastro import repack_h5
from crowdastro import test
from crowdastro import train
from crowdastro import train_cnn


def main():
    parser = argparse.ArgumentParser(description=__description__)
    parser.add_argument(
        '--verbose', '--v', '-v', action='store_true', help='verbose output')
    parser.add_argument(
        '--version', action='store_true', help='get version number')
    subparsers = parser.add_subparsers(dest='subcommand')

    parser_compile_cnn = subparsers.add_parser(
        'compile_cnn', help='compile a convolutional neural network')
    compile_cnn._populate_parser(parser_compile_cnn)

    parser_consensuses = subparsers.add_parser(
        'consensuses',
        help='generate Radio Galaxy Zoo consensus classifications')
    consensuses._populate_parser(parser_consensuses)

    parser_generate_annotator_labels = subparsers.add_parser(
        'generate_annotator_labels',
        help='generates individual annotator labels')
    generate_annotator_labels._populate_parser(parser_generate_annotator_labels)

    parser_generate_cnn_outputs = subparsers.add_parser(
        'generate_cnn_outputs',
        help='generate convolutional neural network training outputs')
    generate_cnn_outputs._populate_parser(parser_generate_cnn_outputs)

    parser_generate_dataset = subparsers.add_parser(
        'generate_dataset', help='generate crowdastro dataset')
    generate_dataset._populate_parser(parser_generate_dataset)

    parser_generate_test_sets = subparsers.add_parser(
        'generate_test_sets', help='generate crowdastro galaxy test sets')
    generate_test_sets._populate_parser(parser_generate_test_sets)

    parser_generate_training_data = subparsers.add_parser(
        'generate_training_data',
        help='generate crowdastro galaxy training data')
    generate_training_data._populate_parser(parser_generate_training_data)

    parser_import_data = subparsers.add_parser(
        'import_data', help='import data into crowdastro')
    import_data._populate_parser(parser_import_data)

    parser_repack_h5 = subparsers.add_parser(
        'repack_h5', help='repacks an HDF5 file')
    repack_h5._populate_parser(parser_repack_h5)

    parser_test = subparsers.add_parser(
        'test', help='tests classifiers')
    test._populate_parser(parser_test)

    parser_train = subparsers.add_parser(
        'train', help='trains classifiers')
    train._populate_parser(parser_train)

    parser_train_cnn = subparsers.add_parser(
        'train_cnn', help='trains a convolutional neural network')
    train_cnn._populate_parser(parser_train_cnn)

    # http://stackoverflow.com/a/11287731/1105803
    if len(sys.argv) < 2:
        sys.argv.append('--help')

    args = parser.parse_args()

    logging.captureWarnings(True)
    if args.verbose:
        logging.root.setLevel(logging.DEBUG)

    if args.version:
        print(__version__)
        return

    subcommands = {
        'compile_cnn': compile_cnn._main,
        'consensuses': consensuses._main,
        'generate_annotator_labels': generate_annotator_labels._main,
        'generate_cnn_outputs': generate_cnn_outputs._main,
        'generate_dataset': generate_dataset._main,
        'generate_test_sets': generate_test_sets._main,
        'generate_training_data': generate_training_data._main,
        'import_data': import_data._main,
        'repack_h5': repack_h5._main,
        'test': test._main,
        'train': train._main,
        'train_cnn': train_cnn._main,
    }

    subcommands[args.subcommand](args)


if __name__ == '__main__':
    main()



"""Trains a convolutional neural network.

Matthew Alger
The Australian National University
2016
"""

import argparse
import logging
import subprocess

import h5py
import numpy

from .config import config

PATCH_RADIUS = config['patch_radius']
PATCH_DIAMETER = PATCH_RADIUS * 2


def train(training_h5, model_json, weights_path, epochs, batch_size, s3=False,
          s3_bucket=None, using_dataset=False, cnn_train_set_h5=None):
    """Trains a CNN.

    training_h5: Training HDF5 file.
    model_json: JSON model file.
    weights_path: Output weights HDF5 file.
    epochs: Number of training epochs.
    batch_size: Batch size.
    s3: Whether to periodically dump to Amazon S3. Default False.
    s3_bucket: Name of the bucket to dump to. Must be specified iff s3 is True.
    using_dataset: Whether the given training file is the Zenodo crowdastro
        dataset. Default False (i.e. we are using the generated training file).
    cnn_train_set_h5: HDF5 file specifying the CNN training set. Overrides the
        training file if specified. Default None (i.e. not specified).
    """
    if s3 and not s3_bucket:
        raise ValueError('Must specify s3_bucket to dump to S3.')
    if not s3 and s3_bucket:
        raise ValueError('s3_bucket was specified but s3 is False.')

    import keras.callbacks
    import keras.models
    model = keras.models.model_from_json(model_json.read())
    model.compile(loss='binary_crossentropy', optimizer='adadelta')

    if not cnn_train_set_h5:
        train_set = training_h5['cnn_train_set'].value
    else:
        train_set = cnn_train_set_h5['cnn_train_set'].value

    if not using_dataset:
        ir_survey = training_h5.attrs['ir_survey']
        n_nonimage_features = config['surveys'][ir_survey]['n_features']
        features_name = 'raw_features'
    else:
        ir_survey = 'wise'
        n_nonimage_features = 5
        features_name = 'features'

    training_inputs = training_h5[features_name].value[
        train_set, n_nonimage_features:]

    training_inputs = training_inputs.reshape(
            (-1, 1, PATCH_DIAMETER, PATCH_DIAMETER))
    training_outputs = training_h5['labels'].value[train_set]
    assert training_inputs.shape[0] == training_outputs.shape[0]

    # Downsample for class balance.
    zero_indices = (training_outputs == 0).nonzero()[0]
    one_indices = (training_outputs == 1).nonzero()[0]
    subset_zero_indices = numpy.random.choice(
        zero_indices, size=(len(one_indices,)), replace=False)
    all_indices = numpy.hstack([subset_zero_indices, one_indices])
    all_indices.sort()

    training_inputs = training_inputs[all_indices]
    training_outputs = training_outputs[all_indices]
    assert (training_outputs == 1).sum() == (training_outputs == 0).sum()

    class DumpToS3(keras.callbacks.Callback):
        def __init__(self, weights_path, bucket, period=50):
            super().__init__()
            self.weights_path = weights_path
            self.bucket = bucket
            self.period = period

        def on_train_begin(self, logs={}):
            self.epochs = 0

        def on_epoch_end(self, epoch, logs={}):
            self.epochs += 1
            if self.epochs % self.period == 0:
                # Every 50 epochs...
                logging.debug('Dumping to S3...')
                res = subprocess.check_call(
                    ['aws', 's3', 'cp', self.weights_path,
                     's3://' + self.bucket + '/', '--region', 'us-east-1'])
                logging.info('Dumped to S3: {}'.format(res))

    try:
        model.load_weights(weights_path)
        logging.info('Loaded weights.')
    except OSError:
        logging.warning('Couldn\'t load weights file. Creating new file...')
        pass

    callbacks = [
        keras.callbacks.ModelCheckpoint(
                weights_path,
                monitor='val_loss',
                verbose=1,
                save_best_only=False,
                save_weights_only=True,
                mode='auto'),
    ]

    if s3:
        callbacks.append(DumpToS3(weights_path, s3_bucket))

    model.fit(training_inputs, training_outputs, batch_size=batch_size,
              nb_epoch=epochs, callbacks=callbacks)
    model.save_weights(weights_path, overwrite=True)


def check_raw_data(training_h5):
    """Sanity check the input data

    training_h5: Training HDF5 file.
    """
    def HDF5_type(name, node):
        if isinstance(node, h5py.Dataset):
            logging.info('Dataset: {}'.format(node.name))
            logging.info('\thas shape {}'.format(node.shape))
        else:
            logging.info('\t{} of type {}'.format((node.name, type(node))))
    logging.info('Peeking into HDF5 file')
    training_h5.visititems(HDF5_type)
    logging.info('End file peeking')


def _populate_parser(parser):
    parser.description = 'Trains a convolutional neural network.'
    parser.add_argument('--h5', default='data/training.h5',
                        help='HDF5 training data file')
    parser.add_argument('--model', default='data/model.json',
                        help='JSON model file')
    parser.add_argument('--output', default='data/weights.h5',
                        help='HDF5 file for output weights')
    parser.add_argument('--epochs', default=10,
                        help='number of epochs to train for')
    parser.add_argument('--batch_size', default=100, help='batch size')
    parser.add_argument('--s3', help='dump to Amazon S3', action='store_true')
    parser.add_argument('--s3_bucket', help='name of S3 bucket', default='')
    parser.add_argument(
        '--cnn_train_set',
        help='HDF5 CNN training set override (default no override)',
        default='')
    parser.add_argument('--dataset', help='training file is the Zenodo dataset',
                        action='store_true')


def _main(args):
    with h5py.File(args.h5, 'r') as training_h5:
        check_raw_data(training_h5)
        with open(args.model, 'r') as model_json:
            if args.cnn_train_set:
                with h5py.File(args.cnn_train_set, 'r') as cnn_train_set_h5:
                    train(training_h5, model_json, args.output,
                          int(args.epochs), int(args.batch_size), s3=args.s3,
                          s3_bucket=args.s3_bucket,
                          cnn_train_set_h5=cnn_train_set_h5,
                          using_dataset=args.dataset)
            else:
                train(training_h5, model_json, args.output,
                      int(args.epochs), int(args.batch_size), s3=args.s3,
                      s3_bucket=args.s3_bucket, using_dataset=args.dataset)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Generates training data (potential hosts and their astronomical features).

Also generates training/testing indices.

Matthew Alger
The Australian National University
2016
"""

import argparse

import h5py
import numpy
import sklearn.neighbors

from .config import config

FITS_HEIGHT = config['surveys']['atlas']['fits_height']
FITS_WIDTH = config['surveys']['atlas']['fits_height']
IMAGE_SIZE = FITS_HEIGHT * FITS_WIDTH  # px
ARCMIN = 1 / 60  # deg


def remove_nans(n):
    """Replaces NaN with 0."""
    if numpy.ma.is_masked(n):
        return 0

    return float(n)


def generate(f_h5, out_f_h5, field='cdfs'):
    """Generates potential hosts and their astronomical features.

    f_h5: crowdastro input HDF5 file.
    out_f_h5: Training data output HDF5 file.
    """
    ir_survey = f_h5.attrs['ir_survey']
    if ir_survey == 'swire':
        swire = f_h5['/swire/{}/numeric'.format(field)]
        fluxes = swire[:, 2:7]
        # Skip stellarities.
        distances = swire[:, 8].reshape((-1, 1))
        images = swire[:, 9:]
        coords = swire[:, :2]
        s1_s2 = fluxes[:, 0] / fluxes[:, 1]
        s2_s3 = fluxes[:, 1] / fluxes[:, 2]
        s3_s4 = fluxes[:, 2] / fluxes[:, 3]
        astro_features = numpy.concatenate(
                [fluxes,
                 s1_s2.reshape((-1, 1)),
                 s2_s3.reshape((-1, 1)),
                 s3_s4.reshape((-1, 1))], axis=1)
    elif ir_survey == 'wise':
        wise = f_h5['/wise/{}/numeric'.format(field)]
        magnitudes = wise[:, 2:6]
        distances = wise[:, 6].reshape((-1, 1))
        images = wise[:, 7:]
        coords = wise[:, :2]

        # Magnitude differences are probably useful features.
        w1_w2 = magnitudes[:, 0] - magnitudes[:, 1]
        w2_w3 = magnitudes[:, 1] - magnitudes[:, 2]

        # Converting the magnitudes to a linear scale seems to improve
        # performance.
        linearised_magnitudes = numpy.power(10, -0.4 * magnitudes)
        w1_w2 = numpy.power(10, -0.4 * w1_w2)
        w2_w3 = numpy.power(10, -0.4 * w2_w3)
        astro_features = numpy.concatenate(
                [linearised_magnitudes,
                 w1_w2.reshape((-1, 1)),
                 w2_w3.reshape((-1, 1))], axis=1)

    n_features = config['surveys'][ir_survey]['n_features']
    assert astro_features.shape[1] + distances.shape[1] == n_features

    # We now need to find the labels for each.
    if field == 'cdfs':
      truths = set(f_h5['/atlas/cdfs/consensus_objects'][:, 1])
      labels = numpy.array([o in truths for o in range(len(astro_features))])

      assert len(labels) == len(astro_features)
      out_f_h5.create_dataset('labels', data=labels)

    assert len(astro_features) == len(distances)
    assert len(distances) == len(images)

    features = numpy.hstack([astro_features, distances, images])
    n_astro = features.shape[1] - images.shape[1]

    # Save to HDF5.
    out_f_h5.create_dataset('raw_features', data=features)
    out_f_h5.create_dataset('positions', data=coords)
    out_f_h5.attrs['ir_survey'] = ir_survey
    out_f_h5.attrs['field'] = field

def _populate_parser(parser):
    parser.description = 'Generates training data (potential hosts and their ' \
                         'astronomical features).'
    parser.add_argument('-i', default='data/crowdastro.h5',
                        help='HDF5 input file')
    parser.add_argument('-o', default='data/training.h5',
                        help='HDF5 output file')
    parser.add_argument('--field', default='cdfs',
                        help='ATLAS field')

def _main(args):
    with h5py.File(args.i, 'r') as f_h5:
        assert f_h5.attrs['version'] == '0.5.1'
        with h5py.File(args.o, 'w') as out_f_h5:
            generate(f_h5, out_f_h5,field=args.field)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Compute checksums for a list of files.

Matthew Alger
The Australian National University
2016
"""

import argparse
import csv
import io

from crowdastro.import_data import hash_file


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filenames', nargs='+')

    args = parser.parse_args()

    out = io.StringIO()
    writer = csv.writer(out)

    for filename in args.filenames:
        with open(filename, 'rb') as f:
            h = hash_file(f)
            writer.writerow((filename, h))

    print(out.getvalue())


if __name__ == '__main__':
    main()
